#! /usr/bin/env python

from __future__ import with_statement

import getopt
import gtp
import re
import sys


def aei_side_to_gtp (side):
    return 'white' if side == 'g' else 'black'

def aei_setup_to_gtp (setup):
    return setup[1:-1].replace (' ', '-')

def aei_move_to_gtp (move):
    return move.replace (' ', '-')

def gtp_move_to_aei (move):
    return move.replace ('-', ' ')


class CleanExit (Exception):
    pass


class BotBridge (object):

    def __init__(self, engine, show_game):
        self.engine      = engine
        self.show_game   = show_game
        self.__options   = { }
        self.__last_move = None


    def run (self):
        try:
            return self._do_run ()
        except CleanExit:
            return 0
        except gtp.UnexpectedFailure, failure:
            response = failure.args[0]
            sys.stderr.write ("The engine returned a failure response to command '%s':\n%s\n"
                              % (response.command, response.text))
            return 1


    def _do_run (self):
        while True:
            line = sys.stdin.readline ()
            if not line:
                return 0

            command, ignored, arguments = line.strip ().partition (' ')
            handler                     = self._handlers.get (command)

            if handler is None:
                sys.stderr.write ("Unknown AEI command '%s'\n" % command)
                return 1

            handler (self, command, arguments)
            sys.stdout.flush ()

    def _aei (self, command, arguments):
        print ('protocol-version 1')
        print ('id name %s' % self.engine.command ('name').assert_success.text_line)
        print ('id version %s' % self.engine.command ('version').assert_success.text_line)
        print ('id author %s' % "Paul Pogonyshev")  # FIXME: Hardcoded
        print ('aeiok')

    def _isready (self, command, arguments):
        print ('readyok')

    def _newgame (self, command, arguments):
        self.engine.clear_board ()
        self.engine.command ('clear_metadata')
        self.__options.clear ()

    def _setposition (self, command, arguments):
        side, setup = arguments.split (None, 1)
        self.engine.command ('setup_squares', aei_side_to_gtp (side), aei_setup_to_gtp (setup))

    def _setoption (self, command, arguments):
        arguments = arguments.split (None, 3)
        if len (arguments) == 4:
            self._do_set_option (arguments[1], arguments[3])
        else:
            self._do_delete_option (arguments[1])

    def _do_set_option (self, name, value):
        if name == 'opponent':
            self.engine.command ('set_metadata', name, value)
        else:
            self.__options[name] = value

    def _do_delete_option (self, name):
        del self.__options[name]
        

    def _makemove (self, command, arguments):
        color            = self._side_to_move ()
        move             = aei_move_to_gtp (arguments)
        self.__last_move = (color, move)

        self.engine.command ('play', color, move)

    def _go (self, command, arguments):
        if arguments == 'ponder':
            # FIXME: Not implemented.
            return

        color = self._side_to_move ()

        if self.show_game and self.__last_move is not None and self.__last_move[0] != color:
            sys.stderr.write ('\nOpponent move: %s\n' % self.__last_move[1])

        self.engine.command ('arimaa_time_settings',
                             self.__options.get ('tcmove',     '0'),
                             self.__options.get ('tcreserve',  '0'),
                             str (float (self.__options.get ('tcpercent', '100')) / 100.0),
                             self.__options.get ('tcmax',      '0'),
                             self.__options.get ('tctotal',    '0'),
                             self.__options.get ('tcturntime', '0'))
        self.engine.command ('arimaa_time_left',
                             color,
                             self.__options.get ('greserve' if color == 'white' else 'sreserve',
                                                 '0'))

        if self.show_game:
            self.echo_board ()

        self._report_best_move (self.engine.command ('genmove', color).assert_success.text_line)

        if self.show_game:
            sys.stderr.write ('\nGenerated move: %s\n' % self.__best_move)
            self.echo_board ()

        # GTP autoplays the generated move, but AEI doesn't.
        self.engine.command ('undo').assert_success


    def _stop (self, command, arguments):
        self._report_best_move ()

    def _quit (self, command, arguments):
        raise CleanExit


    def _side_to_move (self):
        return self.engine.command ('side_to_move').assert_success.text_line

    def _report_best_move (self, best_move = None):
        if best_move is not None:
            self.__best_move = best_move

        print ('bestmove %s' % gtp_move_to_aei (self.__best_move))


    def echo_board (self):
        response = self.engine.command ('showboard').assert_success
        sys.stderr.write ("%s\n\n" % response.text)


    _handlers = { 'aei':         _aei,
                  'isready':     _isready,
                  'newgame':     _newgame,
                  'setposition': _setposition,
                  'setoption':   _setoption,
                  'makemove':    _makemove,
                  'go':          _go,
                  'stop':        _stop,
                  'quit':        _quit }



HELP_STRING = """\
Usage: %s [OPTION]...

Options:
  --command=COMMAND-LINE  run the engine (actual bot) in GTP mode
  --show-game             print position before and after asking the bot to
                          generate a move and the move string

'--command' option is required.  Note that shell-style parsing is
applied to it in order to split the string into argument list.
"""


def main ():
    options, args = getopt.gnu_getopt (sys.argv, '', ['command=', 'show-game'])
    option_dict   = dict (options)

    if not option_dict.get ('--command') or len (args) != 1:
        sys.exit (HELP_STRING % sys.argv[0])

    with gtp.Engine.spawn_as_shell (option_dict['--command']) as engine:
        sys.exit (BotBridge (engine, '--show-game' in option_dict).run ())


try:
    main ()
except KeyboardInterrupt:
    # That's OK.
    pass



# Local variables:
# mode: python
# python-indent: 4
# indent-tabs-mode: nil
# fill-column: 90
# End:
