#! /usr/bin/env python

from __future__ import with_statement

import getopt
import gtp
import re
import sys


_COLOR_TO_PLAY_LINE = re.compile (r'^\d+([wbgs])$')
_MOVE_LINE          = re.compile (r'^\d+([wbgs])\s+(.+)$')


def other_color (color):
    return 'black' if color == 'white' else 'white'

def parse_color (color):
    if color in ('w', 'g'):
        return 'white'
    if color in ('b', 's'):
        return 'black'
    raise ValueError ("failed to parse '%s' as a color" % color)

def parse_color_to_play_line (move_line):
    match = _COLOR_TO_PLAY_LINE.match (move_line)
    if match:
        return parse_color (match.group (1))
    else:
        return None

def parse_move_line (move_line):
    move_line = move_line.rstrip ()
    match     = _MOVE_LINE.match (move_line)

    if match:
        return parse_color (match.group (1)), match.group (2).strip ().replace (' ', '-')
    else:
        raise ValueError ("failed to parse '%s' as a move line" % move_line)


def parse_game_state_line (line):
    line      = line.rstrip ('\n\r').replace ('%13', '\n').replace ('%25', '%')
    key_value = line.split ('=', 1)

    if len (key_value) == 2:
        return key_value[0], key_value[1]
    else:
        return None, None

def parse_game_state_file (game_state_file_name):
    game_state = { }

    with open (game_state_file_name) as game_state_file:
        for line in game_state_file:
            if line:
                key, value = parse_game_state_line (line)
                if key is not None:
                    game_state[key] = value

    return game_state

def get_player_name (game_state, color):
    name = game_state.get ('wplayer' if color == 'white' else 'bplayer')

    if name is not None:
        name = name.strip ()
        if name.startswith ('*'):
            name = name[1:].strip ()

    return name



class BotBridge (object):

    def __init__(self, engine, show_game,
                 position_file_name, moves_file_name, game_state_file_name):
        self.engine               = engine
        self.show_game            = show_game
        self.position_file_name   = position_file_name
        self.moves_file_name      = moves_file_name
        self.game_state_file_name = game_state_file_name


    def run (self):
        try:
            return self._do_run ()
        except gtp.UnexpectedFailure, failure:
            response = failure.args[0]
            sys.stderr.write ("The engine returned a failure response to command '%s':\n%s\n"
                              % (response.command, response.text))
            return 1


    def _do_run (self):
        color_to_play = None
        last_move     = None

        with open (self.moves_file_name) as moves:
            for move_line in moves:
                if not move_line:
                    continue

                # These complications are only needed because of the ancient 'match'
                # script which doesn't set many entries in game state file properly.
                color_to_play = parse_color_to_play_line (move_line)
                if color_to_play is not None:
                    break

                color, move = parse_move_line (move_line)
                engine.command ('play', color, move).assert_success

                if color != color_to_play:
                    last_move = (color, move)

        if self.show_game and last_move is not None and last_move[0] != color_to_play:
            sys.stderr.write ('\nOpponent move: %s\n' % last_move[1])

        game_state = parse_game_state_file (self.game_state_file_name)

        if self.engine.known_command ('set_metadata'):
            # According to Janzert it is not possible to know "real" game id (yet)...
            game_id = game_state.get ('tableId')
            if game_id is not None:
                self.engine.command ('set_metadata', 'game_id', game_id)

            own_name = get_player_name (game_state, color_to_play)
            if own_name is not None:
                self.engine.command ('set_metadata', 'own_name', own_name)

            opponent = get_player_name (game_state, other_color (color_to_play))
            if opponent is not None:
                self.engine.command ('set_metadata', 'opponent', opponent)

            move_number = game_state.get ('plycount')
            if move_number is not None:
                self.engine.command ('set_metadata', 'move_number', move_number)

        self.engine.command ('arimaa_time_settings',
                             game_state['tcmove'],
                             game_state['tcreserve'],
                             str (int (game_state['tcpercent']) / 100.0),
                             game_state['tcmax'],
                             game_state['tctotal'],
                             game_state['tcturntime']).assert_success

        if 'tcwreserve2' in game_state and 'tcbreserve2' in game_state:
            self.engine.command ('arimaa_time_left', 'white',
                                 game_state['tcwreserve2']).assert_success
            self.engine.command ('arimaa_time_left', 'black',
                                 game_state['tcbreserve2']).assert_success

        if self.show_game:
            self.echo_board ()

        response = self.engine.command ('genmove', color_to_play).assert_success

        if self.show_game:
            sys.stderr.write ('\nGenerated move: %s\n' % response.text_line)
            self.echo_board ()

        print (response.text_line.replace ('-', ' '))


    def echo_board (self):
        response = self.engine.command ('showboard').assert_success
        sys.stderr.write ("%s\n\n" % response.text)



options, args = getopt.gnu_getopt (sys.argv, '', ['command=', 'show-game'])
option_dict   = { }

for name, value in options:
    option_dict[name] = value

HELP_STRING = """\
Usage: %s [OPTION]... POS-FILE MOVES-FILE STATE-FILE

Options:
  --command=COMMAND-LINE  run the engine (actual bot) in GTP mode
  --show-game             print position before and after asking the bot to
                          generate a move and the move string

'--command' option is required.  Note that shell-style parsing is
applied to it in order to split the string into argument list.
"""

if not option_dict.get ('--command') or len (args) != 4:
    sys.exit (HELP_STRING % sys.argv[0])

with gtp.Engine.spawn_as_shell (option_dict['--command']) as engine:
    sys.exit (BotBridge (engine, '--show-game' in option_dict, *args[1:]).run ())



# Local variables:
# mode: python
# python-indent: 4
# indent-tabs-mode: nil
# fill-column: 90
# End:
