import re
import subprocess


__all__ = ('Engine', 'Response',
           'ID_REGEXP', 'STEP_REGEXP_STRING', 'MOVE_REGEXP_STRING', 'MOVE_REGEXP')


ID_REGEXP          = re.compile ('\\d+')
STEP_REGEXP_STRING = '[RCDHMErcdhme][a-h][1-8][nwsex]'
MOVE_REGEXP_STRING = '%s(?:-%s)*' % (STEP_REGEXP_STRING, STEP_REGEXP_STRING)
MOVE_REGEXP        = re.compile (MOVE_REGEXP_STRING)


class UnexpectedFailure (Exception):

    def __str__(self):
        response = self.args[0]
        return ("<unexpected failure response to command '%s': %s>"
                % (response.command, response.text.replace ('\n', '\\n')))


class Engine (object):

    def __init__(self, commands, responses, popen = None, initialize_engine = True):
        self.__commands  = commands
        self.__responses = responses
        self.__popen     = popen

        if initialize_engine:
            self.reset ()

    @staticmethod
    def spawn (*command_line):
        popen = subprocess.Popen (command_line,
                                  stdin  = subprocess.PIPE,
                                  stdout = subprocess.PIPE)
        return Engine (popen.stdin, popen.stdout, popen)

    @staticmethod
    def spawn_as_shell (command_line):
        popen = subprocess.Popen (command_line,
                                  shell  = True,
                                  stdin  = subprocess.PIPE,
                                  stdout = subprocess.PIPE)
        return Engine (popen.stdin, popen.stdout, popen)


    def reset (self):
        self.clear_board ()


    def command (self, *arguments):
        response = self.maybe_command (*arguments)
        if response is not None:
            return response
        else:
            raise ValueError ("expected a GTP command, not '%s'" % ' '
                              .join (str (value) for value in arguments))

    def maybe_command (self, *arguments):
        command_line   = ' '.join (str (value) for value in arguments)
        is_command, id = self.get_command_data (command_line)

        self.__commands.write ('%s\n' % command_line.rstrip ('\r\n'))
        self.__commands.flush ()
        if not is_command:
            return None

        response_lines = []
        while True:
            response_line = self.__responses.readline ().rstrip ('\r\n')
            if response_line:
                response_lines.append (response_line.rstrip ())
            else:
                break

        if not response_lines:
            raise ValueError ("empty response to command '%s'" % command_line.strip ())

        if id is None:
            prefix = ' '
        else:
            prefix = '%d ' % id

        for flag in  ('=', '?'):
            prefix = (flag if id is None else '%s%d' % (flag, id))

            if response_lines[0] == prefix:
                response_lines[0] = response_lines[0][len (prefix):]
                return Response (id, command_line, flag == '=', response_lines)

            prefix += ' '
            if response_lines[0].startswith (prefix):
                response_lines[0] = response_lines[0][len (prefix):]
                return Response (id, command_line, flag == '=', response_lines)

        raise ValueError ("unparsable response to command '%s':\n%s"
                          % (command_line.strip (), '\n'.join (response_lines)))


    def known_command (self, command):
        response = self.command ('known_command', command)
        return response.success and response.boolean

    def clear_board (self):
        response = self.command ('clear_board')
        if not response.success:
            raise ValueError ("engine failed to clear the board: %s" % response.text)


    def close (self):
        self.__commands .close ()
        self.__responses.close ()

        if self.__popen is not None:
            self.__popen.wait ()            


    @staticmethod
    def get_command_data (command_line):
        comment_start = command_line.find ('#')
        if comment_start >= 0:
            command_line = command_line[:comment_start]

        command_line = command_line.strip ()
        if not command_line:
            return False, None

        id_match = ID_REGEXP.match (command_line)
        if not id_match:
            return True, None
        else:
            if command_line[id_match.end ():].lstrip ():
                return True, int (id_match.group ())
            else:
                return False, None


    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        try:
            self.close ()
        except:
            pass


class Response (object):

    def __init__(self, id, command, success, lines):
        self.__id      = id
        self.__command = command
        self.__success = success
        self.__lines   = list (lines)

        if len (lines) == 1 and not lines[0]:
            self.__lines = ()


    @property
    def assert_success (self):
        if self.success:
            return self
        else:
            raise UnexpectedFailure (self)


    @property
    def id (self):
        return self.__id

    @property
    def command (self):
        return self.__command

    @property
    def success (self):
        return self.__success

    @property
    def num_lines (self):
        return len (self.__lines)

    @property
    def lines (self):
        return self.__lines

    @property
    def text (self):
        return '\n'.join (self.__lines)

    @property
    def text_line (self):
        if not self.__lines:
            return ''
        elif len (self.__lines) == 1:
            return self.__lines[0]
        else:
            raise ValueError ("expected a single-line response")

    @property
    def boolean (self):
        if self.text_line == 'true':
            return True
        elif self.text_line == 'false':
            return False
        else:
            raise ValueError ("expected 'true' or 'false', got '%s'" % self.text_line)

    @property
    def integer (self):
        return int (self.text_line)

    @property
    def real (self):
        return float (self.text_line)



# Local variables:
# mode: python
# python-indent: 4
# indent-tabs-mode: nil
# fill-column: 90
# End:
