#! /usr/bin/env python

#
# twogtp.py
#
# Play two programs against each other using GTP - The Game Text Protocol
#
#
# This version has support for the following games:
#   - go
#   - amazons
#   - loa
#   - arimaa
#
#
# Examples (see the end for command switches to this program):
#
#  ./twogtp.py --white "./gnugo --mode gtp" \
#              --black "./gnugo --mode gtp --level 5"
#
#  ./twogtp.py --game amazons \
#              --white "./molpadia -G -w t2" \
#              --black "./molpadia -G -b p5:t2"
#

from getopt import *
import popen2
import sys
import string


# Set to 1 to get traces of what is happening.
debug = 0


# ================================================================
#                 Definitions of supported games
#


class Game:
    def __init__(self):
        None

    def starter(self):
        return "unknown"


class Go(Game):
    def __init__(self):
        Game.__init__(self)

    def name(self):
        return "go"

    def starter(self):
        return "Black"

    def move_to_sgfmove(self, size, move):
        global debug
    
        move = string.lower(move)
        if move == "pass":
            return ""
        if debug:
            print "Coords: <" + move + ">"
        letter = move[0]
        digits = move[1:]
        if letter > "i":
            sgffirst = chr(ord(letter) - 1)
        else:
            sgffirst = letter
        sgfsecond = chr(ord("a") + int(size) - int(digits))
        return sgffirst + sgfsecond


class Amazons(Game):
    def __init__(self):
        Game.__init__(self)

    def name(self):
        return "amazons"
    
    def starter(self):
        return "White"

    def move_to_sgfmove(self, size, move):
        return string.lower(move)


class Loa(Game):
    def __init__(self):
        Game.__init__(self)

    def name(self):
        return "loa"
    
    def starter(self):
        return "Black"

    def move_to_sgfmove(self, size, move):
        return string.lower(move)


class Arimaa(Game):
    def __init__(self):
        Game.__init__(self)

    def name(self):
        return "arimaa"
    
    def starter(self):
        return "White"

    def move_to_sgfmove(self, size, move):
        return string.lower(move)



# ================================================================


class GTP_connection:

    #
    # Class members:
    #   outfile		File to write to
    #   infile		File to read from

    def __init__(self, command):
        try:
            print "running: ", command
            infile, outfile = popen2.popen2(command)
        except:
            print "popen2 failed"
            sys.exit(1)
        self.infile  = infile
        self.outfile = outfile
        
    def exec_cmd(self, cmd):
        global debug
        
        if debug:
            sys.stderr.write("GTP command: " + cmd + "\n")
        self.outfile.write(cmd + "\n\n")
        self.outfile.flush()
        
        result = ""
        line = self.infile.readline()
        # A line with just "\n" marks the end of the answer
        while line and (line != "\n"):
            if len(line) > 0:
                # A line starting with % is a trace line.
                if result == "" and line[0] == "%":
                    print line[1:-1]
                else:
                    result = result + line
            line = self.infile.readline()
        if debug:
            sys.stderr.write("Reply: >>>\n" + result + "<<<\n")

        # Remove trailing newline from the result
        if len(result) > 0:
            if result[-1] == "\n":
                result = result[:-1]

        if len(result) == 0:
            return "ERROR: len = 0"
        if (result[0] == "?"):
            #print "ERROR: GTP Command failed: " + result[2:]
            return "ERROR: GTP Command failed: " + result[2:]
        if (result[0] == "="):
            return result[2:]
        return "ERROR: Unrecognized answer: " + result
        

class GTP_player:

    # Class members:
    #    connection		GTP_connection
    #    protocol_version	1 or 2 (only versions supported)

    def __init__(self, command):
        self.connection       = GTP_connection(command)
        self.protocol_version = self.protocol_version()
        if (self.protocol_version != "1"
            and self.protocol_version != "2"):
            print "Unsupported GTP version: ", self.protocol_version
            sys.exit(1)

    #
    # Required commands for GTP version 2
    #
    
    def protocol_version(self):
        return self.connection.exec_cmd("protocol_version")

    def get_program_name(self):
        return self.connection.exec_cmd("name") + " " + \
               self.connection.exec_cmd("version")

    def quit(self):
        return self.connection.exec_cmd("quit")
    
    def boardsize(self, size):
        return self.connection.exec_cmd("boardsize " + size)

    def clear_board(self):
        return self.connection.exec_cmd("clear_board")
    
    def komi(self, komi):
        return self.connection.exec_cmd("komi " + komi)

    def play(self, color, move):
        command = "play " + color + " " + move
        #print "play: command = ", command
        return self.connection.exec_cmd(command)

    def black(self, move):
        if (self.protocol_version == 1):
            return self.connection.exec_cmd("black " + move)
        else:
            return self.play("black", move)

    def white(self, move):
        if (self.protocol_version == 1):
            return self.connection.exec_cmd("white " + move)
        else:
            return self.play("white", move)

    def genmove(self, color):
        if (self.protocol_version == 1):
            command = "genmove_" + color
        else: 
            command = "genmove " + color
        result = self.connection.exec_cmd(command)
        return result

    #
    # Optional commands
    #

    def handicap(self, handi):
        result = self.connection.exec_cmd("fixed_handicap " + handi)
        return string.split(result, " ")

    def showboard(self):
        return self.connection.exec_cmd("showboard")

    def get_random_seed(self):
        result = self.connection.exec_cmd("get_random_seed")
        if result[:5] == "ERROR":
            return "unknown"
        return result

    def estimate_score(self):
        return self.connection.exec_cmd("estimate_score")

    def score(self):
        return self.estimate_score(self)

    def game_is_over(self):
        return int(self.connection.exec_cmd("game_is_over"));

class GTP_go_player(GTP_player):

    def __init__(self, command):
        GTP_player.__init__(self, command)


class GTP_amazons_player(GTP_player):

    def __init__(self, command):
        GTP_player.__init__(self, command)

    def boardsize(self, size):
        None
        
    def handicap(self, handicap):
        None

class GTP_loa_player(GTP_player):

    def __init__(self, command):
        GTP_player.__init__(self, command)

    def boardsize(self, size):
        None
        
    def handicap(self, handicap):
        None


class GTP_arimaa_player(GTP_player):

    def __init__(self, command):
        GTP_player.__init__(self, command)

    def boardsize(self, size):
        None
        
    def handicap(self, handicap):
        None



class GTP_game:

    # Class members:
    #    game            Game
    #    whiteplayer	 GTP_player
    #    blackplayer	 GTP_player
    #    size		 int
    #    komi		 float
    #    handicap        int
    #    handicap_stones int
    #    moves		 list of string
    #    to_play         the side to play next
    #    resultw
    #    resultb

    def __init__(self, game, whitecommand, blackcommand, size, komi, handicap):
        self.game = game
        if game.name() == "go":
            self.whiteplayer = GTP_go_player(whitecommand)
            self.blackplayer = GTP_go_player(blackcommand)
        elif game.name() == "amazons":
            self.whiteplayer = GTP_amazons_player(whitecommand)
            self.blackplayer = GTP_amazons_player(blackcommand)
        elif game.name() == "loa":
            self.whiteplayer = GTP_loa_player(whitecommand)
            self.blackplayer = GTP_loa_player(blackcommand)
        elif game.name() == "arimaa":
            self.whiteplayer = GTP_arimaa_player(whitecommand)
            self.blackplayer = GTP_arimaa_player(blackcommand)
        else:
            None
        
        self.size = size
        self.komi = komi
        self.handicap = handicap

    def writesgf(self, sgffilename):
        "Write the game to an SGF file after a game"

        size = self.size
        outfile = open(sgffilename, "w")
        if not outfile:
            print "Couldn't create " + sgffilename
            return
        black_name = self.blackplayer.get_program_name()
        white_name = self.whiteplayer.get_program_name()
        black_seed = self.blackplayer.get_random_seed()
        white_seed = self.whiteplayer.get_random_seed()
        handicap = self.handicap
        komi     = self.komi
        result   = self.resultw + " " + self.resultb

        outfile.write("(;GM[1]FF[4]RU[Japanese]SZ[%s]HA[%s]KM[%s]RE[%s]\n" %
                      (size, handicap, komi, result))
        outfile.write("PW[%s (random seed %s)]PB[%s (random seed %s)]\n" %
                      (white_name, white_seed, black_name, black_seed))

        if handicap > 1:
            for stone in self.handicap_stones:
                outfile.write("AB[%s]", self.game.move_to_sgfmove(size, stone))
            outfile.write("\n")

        to_play = self.game.starter()
        if self.game.name() == "go":
            if int(handicap) < 2:
                to_play = "B"
            else:
                to_play = "W"

        for move in self.moves:
            sgfmove = self.game.move_to_sgfmove(size, move)
            outfile.write(";%s[%s]\n" % (to_play, sgfmove))
            if to_play == "B":
                to_play = "W"
            else:
                to_play = "B"
        outfile.write(")\n")
        outfile.close

    def game_is_over(self, passes):
        # FIXME: Ugly hack.  Generalize this.
        if self.game.name() == "go":
            return passes >= 2
        else:
            if self.to_play == "B":
                side = self.blackplayer
            else:
                side = self.whiteplayer
            return side.game_is_over();

    def play(self, sgffile):
        "Play a game"
        global verbose

        if verbose >= 1:
            print "Setting boardsize and komi for black\n"
        self.blackplayer.boardsize(self.size)
        self.blackplayer.komi(self.komi)
        self.blackplayer.clear_board()

        if verbose >= 1:
            print "Setting boardsize and komi for white\n"
        self.whiteplayer.boardsize(self.size)
        self.whiteplayer.komi(self.komi)
        self.whiteplayer.clear_board()

        self.handicap_stones = []
        self.to_play = self.game.starter()
        print self.to_play, "starts the game"
        self.to_play = self.to_play[0]

        # FIXME: Ugly hack.  Genaralize this.
        if self.game.name() == "go":
            if int(self.handicap) < 2:
                self.to_play = "B"
            else:
                self.handicap_stones = self.whiteplayer.handicap(self.handicap)
                for stone in self.handicap_stones:
                    self.blackplayer.black(stone)
                self.to_play = "W"

        # Play the actual game.
        self.moves = []
        passes = 0
        while not self.game_is_over(passes):
            if self.to_play == "B":
                move = self.blackplayer.genmove("black")
                if move[:5] == "ERROR":
                    # FIXME: write_sgf
                    sys.exit(1)

                self.moves.append(move)
                if string.lower(move[:4]) == "pass":
                    passes = passes + 1
                else:
                    passes = 0
                    result = self.whiteplayer.black(move)
                    if result and result[:5] == "ERROR":
                        print "Error in game on white move", move, ":", result
                        sys.exit(1)
                    if verbose >= 1:
                        print "Black plays " + move

                self.to_play = "W"
            else:
                move = self.whiteplayer.genmove("white")
                if move[:5] == "ERROR":
                    # FIXME: write_sgf
                    sys.exit(1)

                self.moves.append(move)
                if string.lower(move[:4]) == "pass":
                    passes = passes + 1
                else:
                    passes = 0
                    result = self.blackplayer.white(move)
                    if result and result[:5] == "ERROR":
                        print "Error in game:", result
                        sys.exit(1)
                    if verbose >= 1:
                        print "White plays " + move
                self.to_play = "B"
            if verbose >= 2:
                # print self.blackplayer.showboard()
                print self.whiteplayer.showboard()
                sys.stdout.flush()

        self.resultw = self.whiteplayer.estimate_score()
        self.resultb = self.blackplayer.estimate_score()
        if self.resultb == self.resultw:
            print "Result: ", self.resultw
        else:
            print "Result according to W: ", self.resultw
            print "Result according to B: ", self.resultb
        if sgffile != "":
            self.writesgf(sgffile)

    def result(self):
        return (self.resultw, self.resultb)
    
    def quit(self):
        self.blackplayer.quit()
        self.whiteplayer.quit()


class GTP_match:

    # Class members:
    #    game
    #    black
    #    white
    #    size
    #    komi
    #    handicap

    def __init__(self, game, whitecommand, blackcommand, size, komi, handicap):
        self.game  = game
        self.white = whitecommand
        self.black = blackcommand
        self.size  = size
        self.komi  = komi
        self.handicap = handicap

    def play(self, games, sgfbase):
        game = GTP_game(self.game, self.white, self.black,
                        self.size, self.komi, self.handicap)
        results = []
        for i in range(int(games)):
            sgffilename = "%s%03d.sgf" % (sgfbase, i + 1)
            print "Game ", i + 1
            game.play(sgffilename)
            result = game.result()
            print "Result = ", result
            results.append(result)
        game.quit()
        return results


# ================================================================
#                      Main program
#


# Default values
#
game     = Arimaa()
white    = ""
black    = ""
komi     = "5.5"
size     = "8"
handicap = "0"

games   = 1
sgfbase = "twogtp"

verbose = 0

helpstring = """

Run with:

twogtp [--game go | amazons | loa | arimaa] \\
       --white \'<path to program 1> --mode gtp [program options]\' \\
       --black \'<path to program 2> --mode gtp [program options]\' \\
       [twogtp options]

Possible twogtp options:

  --verbose 1 (to list moves) or --verbose 2 (to draw board)
  --komi <amount>
  --handicap <amount>
  --size <board size>               (default 10)
  --games <number of games to play> (default 1)
  --sgfbase <filename>              (create sgf files with sgfbase as basename)
"""

def usage():
    print helpstring
    sys.exit(1)

(opts, params) = getopt(sys.argv[1:], "",
                        ["game=",
                         "black=",
                         "white=",
                         "verbose=",
                         "komi=",
                         "boardsize=", "size=",
                         "handicap=",
                         "games=",
                         "sgfbase=",
                         ])
for opt, value in opts:
    if opt == "--game":
        if value == "go":
            # the default
            None
        elif value == "amazons":
            game = Amazons()
        elif value == "loa":
            game = Loa()
        elif value == "arimaa":
            game = Arimaa()

    elif opt == "--black":
        black = value
    elif opt == "--white":
        white = value
    elif opt == "--verbose":
        verbose = int(value)
    elif opt == "--komi":
        komi = value
    elif opt == "--boardsize" or opt == "--size":
        size = value
    elif opt == "--handicap":
        handicap = value
    elif opt == "--games":
        games = int(value)
    elif opt == "--sgfbase":
        sgfbase = value
        
if params != []:
    usage()
            

if black == "" or white == "":
    usage()

if (int(handicap) > 1):
    komi = "0.5"

match = GTP_match(game, white, black, size, komi, handicap)
results = match.play(games, sgfbase)

i = 0
for resw, resb in results:
    i = i + 1
    if resw == resb:
        print "Game %d: %s" % (i, resw)
    else:
        print "Game %d: %s %s" % (i, resb, resw)

